import { ForbiddenRpcException } from '@freeblox/shared';
import { Inject, Injectable } from '@nestjs/common';
import { ConfigService } from '@nestjs/config';
import { EncryptJWT, JWK, KeyLike, SignJWT, jwtDecrypt, jwtVerify } from 'jose';

@Injectable()
export class JWTService {
  constructor(
    @Inject('APP_PRIVATE_KEY') private readonly privateKey: KeyLike,
    @Inject('APP_SECRET_KEY') private readonly secretKey: KeyLike,
    @Inject('APP_PUBLIC_KEY') private readonly publicKey: KeyLike,
    @Inject('APP_PUBLIC_KEY_JWK') private readonly publicKeyJWK: JWK,
    private readonly config: ConfigService,
  ) {}

  async sign(data: Record<string, unknown>, audience = 'urn:freeblox:service') {
    const alg = this.config.get('security.jwtAlgorithm');
    const exp =
      this.config.get('security.jwtTokenExpiry') +
      Math.floor(Date.now() / 1000);
    const jwt = await new SignJWT(data)
      .setProtectedHeader({ alg })
      .setIssuedAt()
      .setIssuer('urn:freeblox:auth')
      .setAudience(audience)
      .setExpirationTime(exp)
      .sign(this.privateKey);
    return jwt;
  }

  async verify(jwt: string, audience = 'urn:freeblox:service') {
    const alg = this.config.get('security.jwtAlgorithm');
    const { payload, protectedHeader } = await jwtVerify(jwt, this.publicKey, {
      issuer: 'urn:freeblox:auth',
      audience,
    });
    if (protectedHeader.alg !== alg)
      throw new ForbiddenRpcException('Provided JWT contains invalid headers.');
    return payload;
  }

  async encrypt(
    data: Record<string, unknown>,
    audience = 'urn:freeblox:service',
  ) {
    const alg = this.config.get('security.secretAlgorithm');
    const exp =
      this.config.get('security.refreshTokenExpiry') +
      Math.floor(Date.now() / 1000);
    const jwt = await new EncryptJWT(data)
      .setProtectedHeader({ alg: 'dir', enc: alg })
      .setIssuedAt()
      .setIssuer('urn:freeblox:auth')
      .setAudience(audience)
      .setExpirationTime(exp)
      .encrypt(this.secretKey);
    return jwt;
  }

  async decrypt(jwt: string, audience = 'urn:freeblox:service') {
    const alg = this.config.get('security.secretAlgorithm');
    const { payload, protectedHeader } = await jwtDecrypt(jwt, this.secretKey, {
      issuer: 'urn:freeblox:auth',
      audience,
    });
    if (protectedHeader.enc !== alg)
      throw new ForbiddenRpcException('Provided JWT contains invalid headers.');
    return payload;
  }
}
