import { Injectable } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { UserTokenEntity } from '../database/entities/user-token.entity';
import { JWTService } from './jwt.service';
import { UserEntity } from '../database/entities/user.entity';
import { Repository } from 'typeorm';
import {
  UnauthorizedRpcException,
  UserTokenType,
  generateString,
} from '@freeblox/shared';
import { ConfigService } from '@nestjs/config';

@Injectable()
export class RefreshService {
  constructor(
    @InjectRepository(UserTokenEntity)
    private readonly userTokenRepository: Repository<UserTokenEntity>,
    private readonly jwt: JWTService,
    private readonly config: ConfigService,
  ) {}

  /**
   * Find new refresh token for old token
   * @param token Old token
   * @returns New Token
   */
  async findNewIteration(token: UserTokenEntity) {
    return this.userTokenRepository.findOne({
      where: {
        previous: { id: token.id },
        type: UserTokenType.REFRESH,
      },
    });
  }

  /**
   * Consume a refresh token.
   * @param token Refresh token
   * @returns New refresh token
   */
  async useRefreshToken(token: string) {
    if (!token) throw new UnauthorizedRpcException('Invalid refresh token');
    const decrypted = await this.jwt.decrypt(token);

    if (!decrypted.token || !decrypted.sub)
      throw new UnauthorizedRpcException('Invalid refresh token');

    const tokenEntity = await this.userTokenRepository.findOneOrFail({
      where: {
        token: decrypted.token as string,
        type: UserTokenType.REFRESH,
      },
      relations: ['user'],
    });

    if (decrypted.sub !== tokenEntity.user.id)
      throw new UnauthorizedRpcException('Invalid refresh token');

    // Using an expired refresh token
    if (tokenEntity.expiresAt.getTime() < Date.now()) {
      const newIteration = await this.findNewIteration(tokenEntity);
      // We have already issued a new refresh token, this is probably stolen
      // ..so we delete the whole tree.
      if (newIteration) {
        await this.userTokenRepository.remove(tokenEntity);
      }

      throw new UnauthorizedRpcException('Invalid refresh token');
    }

    // Mark old token as expired
    await this.userTokenRepository.update(
      { id: tokenEntity.id },
      { expiresAt: new Date() },
    );

    // Issue a new refresh token
    const newToken = await this.issueRefreshToken(
      tokenEntity.user,
      tokenEntity,
    );

    return {
      token: newToken,
      user: tokenEntity.user,
    };
  }

  /**
   * Issue a refresh token.
   * @param user User
   * @param previous Previous refresh token
   * @returns New Refresh token
   */
  async issueRefreshToken(user: UserEntity, previous?: UserTokenEntity) {
    const newRefreshToken = await this.createRefreshToken(user);
    const expiry = new Date(
      Date.now() + this.config.get('security.refreshTokenExpiry') * 1000,
    );
    await this.userTokenRepository.save({
      token: newRefreshToken.token,
      user,
      type: UserTokenType.REFRESH,
      expiresAt: expiry,
      previous,
    });
    return newRefreshToken.encrypted;
  }

  /**
   * Generate refresh token and encrypt it
   * @param user User to issue the token to
   * @returns Encrypted and unencrypted tokens
   */
  private async createRefreshToken(user: UserEntity) {
    const token = generateString(512);
    const encrypted = await this.jwt.encrypt({ sub: user.id, token });
    return {
      encrypted,
      token,
    };
  }
}
