import { Injectable } from '@nestjs/common';
import { InjectEntityManager, InjectRepository } from '@nestjs/typeorm';
import { EntityManager, Repository } from 'typeorm';
import { UserEntity } from '../database/entities/user.entity';
import { RoleEntity } from '../database/entities/role.entity';
import { PrivilegeEntity } from '../database/entities/privilege.entity';

@Injectable()
export class RoleService {
  constructor(
    @InjectEntityManager() private manager: EntityManager,
    @InjectRepository(UserEntity)
    private readonly userRepository: Repository<UserEntity>,
  ) {}

  /**
   * Get all applicable privileges for roles
   * @param roles Roles to look for privileges for
   * @param startingPrivileges Privileges already included
   * @returns Unique list of privileges
   */
  async getApplicablePrivileges(
    roles: RoleEntity[],
    startingPrivileges: PrivilegeEntity[] = [],
  ) {
    const roleTree = await this.manager
      .getTreeRepository(RoleEntity)
      .findTrees({ relations: ['privileges'] });

    const privileges: PrivilegeEntity[] = [...startingPrivileges];
    for (const { id } of roles) {
      for (const role of roleTree) {
        const [result, list] = this.accumulatePrivilegesFromTree(role, id);
        if (!result) continue;
        privileges.push(...list);
        break;
      }
    }

    return privileges.filter(
      (value, index, array) =>
        array.findIndex((entry) => entry.id === value.id) === index,
    );
  }

  /**
   * Get all applicable privileges for user.
   * @param user User
   * @returns Unique privilege list
   */
  async getUserPrivileges(user: UserEntity, exclude: PrivilegeEntity[] = []) {
    if (!user.privileges || !user.roles) {
      user = await this.userRepository.findOne({
        where: { id: user.id },
        relations: ['privileges', 'roles'],
      });
    }

    const applicable = await this.getApplicablePrivileges(
      user.roles,
      user.privileges,
    );

    // Exclude privileges provided
    return applicable.filter(
      (privilege) => !exclude.some((entry) => entry.id === privilege.id),
    );
  }

  /**
   * Accumulate privileges from a tree structure of roles.
   * @param root Node to start from
   * @param target Target role ID
   * @returns `[<Found role entity>, <Privilege list>]`
   */
  private accumulatePrivilegesFromTree(root: RoleEntity, target: number) {
    const privileges: PrivilegeEntity[] = [...root.privileges];
    if (root.id === target) return [root, privileges];
    for (const child of root.children || []) {
      const [found, list] = this.accumulatePrivilegesFromTree(child, target);
      if (found) return [found, [...privileges, ...list]];
      continue;
    }
    return [null, []];
  }
}
