import { MetaEntity } from '@freeblox/shared';
import { Exclude, Expose, Type } from 'class-transformer';
import {
  Column,
  CreateDateColumn,
  Entity,
  JoinColumn,
  ManyToOne,
  OneToMany,
  PrimaryGeneratedColumn,
} from 'typeorm';
import { ContentEntity } from './content.entity';
import { ContentRevisionEntity } from './content-revision.entity';
import { ModeratorAction } from '../../enums/moderation-action.enum';
import { RejectionReason } from '../../enums/rejection-reason.enum';
import {
  IsBoolean,
  IsEnum,
  IsNumber,
  IsOptional,
  IsString,
} from 'class-validator';
import { BanEntity } from 'apps/auth/src/database/entities/ban.entity';

@Entity('content_moderation')
@Expose()
export class ContentModerationEntity extends MetaEntity {
  @PrimaryGeneratedColumn()
  id: number;

  @Column({ name: 'content_id' })
  @IsNumber()
  contentId: number;

  @Column({ name: 'revision_id' })
  @IsNumber()
  revisionId: number;

  @Column({ name: 'user_id', type: 'uuid', nullable: true })
  @Exclude()
  userId: string;

  @Column({
    type: 'enum',
    enum: ModeratorAction,
    default: ModeratorAction.PENDING,
  })
  @IsEnum(ModeratorAction)
  action: ModeratorAction;

  @Column({
    type: 'enum',
    enum: RejectionReason,
    nullable: true,
    name: 'rejection_reason',
  })
  @IsEnum(RejectionReason)
  rejectionReason?: RejectionReason;

  @Column({ nullable: true })
  @IsString()
  @IsOptional()
  description?: string;

  @Column({ default: false })
  @IsBoolean()
  @IsOptional()
  penalty: boolean;

  @Column({ default: false, name: 'asset_delete' })
  @IsBoolean()
  @IsOptional()
  assetDelete: boolean;

  @CreateDateColumn({ name: 'created_at' })
  @Expose()
  createdAt: Date;

  @CreateDateColumn({ name: 'decided_at' })
  decidedAt: Date;

  @ManyToOne(() => ContentEntity, { onDelete: 'CASCADE' })
  @JoinColumn({ name: 'content_id' })
  content: ContentEntity;

  @ManyToOne(() => ContentRevisionEntity, { onDelete: 'CASCADE' })
  @JoinColumn({ name: 'revision_id' })
  revision: ContentRevisionEntity;

  @OneToMany(() => ContentModerationBanEntity, (ban) => ban.moderation)
  @Type(() => ContentModerationBanEntity)
  bans?: ContentModerationBanEntity[];
}

@Entity('content_moderation_ban')
export class ContentModerationBanEntity {
  @PrimaryGeneratedColumn()
  @Expose()
  id: number;

  @Column({ nullable: true, name: 'ban_id' })
  banId: number;

  @ManyToOne(() => ContentModerationEntity, (mod) => mod.bans)
  @JoinColumn({ name: 'content_moderation_id' })
  @Expose()
  moderation: ContentModerationEntity;

  @Expose()
  ban?: BanEntity;
}
