import { Exclude, Expose } from 'class-transformer';
import {
  Column,
  CreateDateColumn,
  Entity,
  Index,
  JoinColumn,
  ManyToOne,
  PrimaryGeneratedColumn,
} from 'typeorm';
import { ContentEntity } from './content.entity';
import { Currency } from '../../enums/currency.enum';
import { UserMetaEntity } from '@freeblox/shared';
import { OwnershipType } from '../../enums/ownership-type.enum';

@Entity('content_ownership')
@Expose()
export class ContentOwnershipEntity extends UserMetaEntity {
  @PrimaryGeneratedColumn()
  id: number;

  @Column({ name: 'content_id' })
  contentId: number;

  @Column({ name: 'user_id', type: 'uuid', nullable: true })
  @Exclude()
  userId: string;

  @Column({ type: 'enum', enum: OwnershipType, default: OwnershipType.AUTHOR })
  source: OwnershipType;

  @Column({ name: 'previous_ownership_id', nullable: true })
  @Exclude()
  previousOwnershipId: string;

  @Column({ name: 'purchase_price', nullable: true })
  @Index()
  purchasePrice: number;

  @Column({
    type: 'enum',
    enum: Currency,
    name: 'purchase_currency',
    nullable: true,
  })
  @Index()
  purchaseCurrency: Currency;

  @Column({ nullable: true })
  @Index()
  serial: number;

  @ManyToOne(() => ContentEntity, { onDelete: 'CASCADE' })
  @JoinColumn({ name: 'content_id' })
  @Exclude()
  content: ContentEntity;

  @ManyToOne(() => ContentOwnershipEntity, { onDelete: 'SET NULL' })
  @JoinColumn({ name: 'previous_ownership_id' })
  @Exclude()
  previous?: ContentOwnershipEntity;

  @CreateDateColumn({ name: 'created_at' })
  @Expose()
  createdAt: Date;

  @Column({ type: Date, name: 'ended_at', nullable: true })
  endedAt: Date;

  @Column({ type: Date, name: 'expires_at', nullable: true })
  expiresAt: Date;

  /**
   * is ownership expired
   */
  get expired() {
    return this.expiresAt.getTime() < Date.now();
  }

  /**
   * is ownership invalid (Expired or ended)
   */
  get invalid() {
    return !!this.endedAt || this.expired;
  }
}
