import {
  NotFoundRpcException,
  PageQuery,
  PreconditionFailedRpcException,
  UserInfo,
} from '@freeblox/shared';
import { Inject, Injectable } from '@nestjs/common';
import { InjectEntityManager } from '@nestjs/typeorm';
import { EntityManager, In } from 'typeorm';
import { ContentEntity } from '../database/entities/content.entity';
import { Privacy } from '../enums/privacy.enum';
import { instanceToPlain, plainToClass } from 'class-transformer';
import { ContentRevisionEntity } from '../database/entities/content-revision.entity';
import { ContentAssetType } from '../enums/content-asset-type.enum';
import { ContentType } from '../enums/content-type.enum';
import { ClientProxy } from '@nestjs/microservices';
import { lastValueFrom } from 'rxjs';
import { ContentUserDto } from '../dtos/content-user.dto';
import { ContentVoteEntity } from '../database/entities/content-vote.entity';
import { Vote } from '../enums/vote.enum';
import { ContentCategoryEntity } from '../database/entities/content-category.entity';

@Injectable()
export class CatalogService {
  constructor(
    @InjectEntityManager() private manager: EntityManager,
    @Inject('auth') private authClient: ClientProxy,
  ) {}

  /**
   * Get the assets for the catalog item.
   * @param itemId Item ID
   * @returns Assets for the latest revision
   */
  async getCatalogItemAssets(
    itemId: number,
    typeFilter?: ContentAssetType[],
    typeNameFilter?: string[],
  ) {
    // Load the latest revision to get the assets
    const latestRevision = await this.manager.findOne(ContentRevisionEntity, {
      where: {
        content: { id: itemId },
        assets:
          typeFilter?.length || typeNameFilter?.length
            ? {
                type: typeFilter?.length ? In(typeFilter) : undefined,
                typeName: typeNameFilter?.length
                  ? In(typeNameFilter)
                  : undefined,
              }
            : undefined,
      },
      order: { createdAt: 'DESC' },
      relations: ['assets'],
    });

    if (!latestRevision) return [];
    return latestRevision.assets;
  }

  /**
   * Get catalog item by ID.
   * @param itemId Item ID
   * @param user User
   * @returns Catalog item
   */
  async getCatalogItemById(itemId: number, user?: UserInfo) {
    const findItem = await this.manager.findOne(ContentEntity, {
      where: { id: itemId, deletedAt: null },
      relations: ['prices'],
    });

    if (!findItem) {
      throw new NotFoundRpcException(
        `Catalog item by ID ${itemId} was not found`,
      );
    }

    // Permission checks
    if (
      user &&
      !user.privileges.includes('root') &&
      !user.privileges.includes('contentedit')
    ) {
      // When private, only creator can access
      // TODO: friends
      if (
        findItem.privacy === Privacy.PRIVATE ||
        findItem.privacy === Privacy.FRIENDS
      ) {
        if (findItem.userId !== user.sub) {
          throw new NotFoundRpcException(
            `Catalog item by ID ${itemId} was not found`,
          );
        }
      }
    }

    // Get the owner user
    findItem.user = await this.getContentUser(findItem.userId);

    return instanceToPlain(findItem);
  }

  /**
   * Get paginated comments for catalog item by ID.
   * @param itemId Catalog item ID
   * @param paging Pagination
   * @param user Requesting user
   * @returns Comments
   */
  async getCatalogItemComments(
    itemId: number,
    paging?: PageQuery,
    user?: UserInfo,
  ) {
    const page = paging.page ?? 0;
    const pageSize = paging.pageSize || 10;

    const catalogItem = await this.getCatalogItemById(itemId, user);
    if (!catalogItem.commentsEnabled) {
      throw new PreconditionFailedRpcException('Comments are disabled');
    }

    const [comments, total] = await this.manager.findAndCount(ContentEntity, {
      where: {
        type: ContentType.COMMENT,
        privacy: Privacy.PUBLIC,
        published: true,
        deletedAt: null,
        parent: { id: itemId },
      },
      order: {
        createdAt: 'DESC',
      },
      take: pageSize,
      skip: page * pageSize,
    });

    const mapped = await this.mapComments(comments, user);
    return {
      pagination: {
        total,
        count: mapped.length,
        page,
        pageSize,
      },
      list: instanceToPlain(mapped),
    };
  }

  /**
   * Get votes for a catalog item.
   * @param itemId Content ID
   * @param user User
   * @returns Votes
   */
  async getContentVotes(itemId: number, user?: UserInfo) {
    await this.getCatalogItemById(itemId, user);

    const voteList = await this.manager.find(ContentVoteEntity, {
      where: { contentId: itemId },
    });

    return {
      up: voteList.filter(({ vote }) => vote === Vote.UPVOTE).length,
      down: voteList.filter(({ vote }) => vote === Vote.DOWNVOTE).length,
    };
  }

  /**
   * Get category tree for catalog
   * @returns Category tree
   */
  async getCategoryTree() {
    const categoryTree = await this.manager
      .getTreeRepository(ContentCategoryEntity)
      .findTrees({ relations: ['contentTypes'] });
    return instanceToPlain(categoryTree);
  }

  /**
   * Get user information for content.
   * @private
   * @param userId User ID
   * @returns User information DTO
   */
  async getContentUser(userId: string) {
    const user = await lastValueFrom(
      this.authClient.send('auth.getUserById', { id: userId }),
    );
    return plainToClass(ContentUserDto, user);
  }

  /**
   * Map comment objects.
   * @private
   * @param comments Comments list
   * @param user User
   * @returns Mapped comments
   */
  async mapComments(comments: ContentEntity[], user?: UserInfo) {
    for (const comment of comments) {
      comment.user = await this.getContentUser(comment.userId);

      if (comment.restricted) {
        if (!user.privileges.includes('contentmod')) {
          comment.description = null;
          comment.name = null;
        }
      }
    }
    return comments;
  }
}
